﻿using System;

//  以下の Nodeクラスは、C# コンパイラと Java コンパイラのどちらでも有効な実装です。 

//==== ==== ★★ノードを表すクラス★★ ==== ==== 
public class Node {
    //メンバ変数(フィールド)
    public int code;   //検索対象となるキー
    public int price;  //その他のデータ
    public Node left = null;
    public Node right = null;
    //コンストラクタ
    public Node(int cd, int pr) {
        code = cd;
        price = pr;
    }
}
//==== ==== ★★ ★★ ==== ==== 


class BSTreeApp
{
    static void Main(string[] args)
    {
        // 追加
        BSTreeAdd(new Node(50, 100));
        BSTreeAdd(new Node(22, 200));
        BSTreeAdd(new Node(72, 300));
        BSTreeAdd(new Node(17, 400));
        BSTreeAdd(new Node(39, 500));
        BSTreeAdd(new Node(48, 600));
        BSTreeAdd(new Node(65, 700));
        BSTreeAdd(new Node(90, 800));
        BSTreeAdd(new Node(80, 900));

        // ダンプ
        for (int i = 1; i <= 4; i++)
        {
            Dump(root, i);
            Console.WriteLine();
        }

        int key;
        Node hit;
        int result;

        // 探索
        key = 48;
        hit = BSTreeSearch(root, key);  //二分探索木による探索(「48」を探す)
        if (hit == null)
            Console.WriteLine("target={0} はありません。", key);
        else
            Console.WriteLine("target={0} の価格は {1}です。", hit.code, hit.price);

        // 削除
        result = BSTreeDelete(50);
        if (result < 0)
            Console.WriteLine("50 の削除に失敗しました。");
        result = BSTreeDelete(79);
        if (result < 0)
            Console.WriteLine("79 の削除に失敗しました。");

        // ダンプ
        for (int i = 1; i <= 4; i++)
        {
            Dump(root, i);
            Console.WriteLine();
        }

        // 探索
        key = 48;
        hit = BSTreeSearch(root, key);  //二分探索木による探索(「48」を探す)
        if (hit == null)
            Console.WriteLine("target={0} はありません。", key);
        else
            Console.WriteLine("target={0} の価格は {1}です。", hit.code, hit.price);

        // 削除
        result = BSTreeDelete(48);
        if (result < 0)
            Console.WriteLine("48 の削除に失敗しました。");

        // ダンプ
        for (int i = 1; i <= 4; i++)
        {
            Dump(root, i);
            Console.WriteLine();
        }

        // 探索
        key = 48;
        hit = BSTreeSearch(root, key);  //二分探索木による探索(「48」を探す)
        if (hit == null)
            Console.WriteLine("target={0} はありません。", key);
        else
            Console.WriteLine("target={0} の価格は {1}です。", hit.code, hit.price);

        // 削除
        BSTreeDelete(22);
        BSTreeDelete(72);
        BSTreeDelete(17);
        BSTreeDelete(39);
        BSTreeDelete(65);
        BSTreeDelete(90);
        BSTreeDelete(80);

        //
        Console.WriteLine(root == null);
    }

    //テスト用階層ダンプ
    private static void Dump(Node node, int level)
    {
        level--;
        if (level == 0)
        {
            Console.Write("{0} ", (node == null) ? "*" : node.code.ToString());
            return;
        }
        else
        {
            if (node == null) return;
            Dump(node.left, level);
            Dump(node.right, level);
        }
    }

    //  以下のコード断片は、C# コンパイラと Java コンパイラのどちらでも有効な実装です。 

    //==== ==== ★★二分探索木による探索本体★★ ==== ==== 
    private static Node root = null;   //ルートノード

    //探索
    //  r ---- 探索の起点ノード
    //  target ---- 検索したいコード(code)
    //  戻り値 ---- ヒットしたノード(該当ない場合はnull)
    private static Node BSTreeSearch(Node root, int target) {
        if (root == null) //ルートが無ければ該当なし
            return null;

        for (Node p = root; p != null; ) { //現在の検索対象ノードを表す p が null になるまで繰り返す
            if (p.code == target)
                return p;           //検出して終了
            else if (p.code > target)
                p = p.left;         //左に進む
            else
                p = p.right;        //右に進む
        }

        return null; //該当なしとして null を返して終了
    }

    //追加
    //  q      ---- 追加したいノード
    //  戻り値 ---- 正常終了なら 0、負の値は何らかのエラーコード
    private static int BSTreeAdd(Node q) {
        if (root == null) { //ルートがなければルートにして終了
            root = q;
            return 0;
        }

        int target = q.code;
        for (Node p = root; ; ) { //空きが見つかるまで進めて追加する
            if (p.code == target)
                return -1;             //重複エラーとしてエラー終了
            else if (p.code > target) {
                if (p.left == null) {  //左の空き確認
                    p.left = q;
                    return 0;
                }
                else
                    p = p.left;        //左に進む
            }
            else {
                if (p.right == null) { //右の空き確認
                    p.right = q;
                    return 0;
                }
                else
                    p = p.right;       //右に進む
            }
        }
    }

    //削除
    //  target ---- 削除したいノードのコード
    //  戻り値 ---- 正常終了なら 0、負の値は何らかのエラーコード
    private static int BSTreeDelete(int target) {
        if (root == null) {  //ルートがなければエラー終了
            return -1;       //該当なし
        }

        if (root.code == target) { //ルートで一致した場合、ルートより上の親の修正は不要なので、単にこのノードから下の修正
            root = DeleteAndNewSubTree(root);
            return 0;
        }

        //ここまで実行できた場合は、rootが削除対象でない

        for (Node parent = root; ; ) { //一致するノードを見つけて削除する、一致を見つけたとき、parentはその親ノード

            //以下の評価では、parent.code == target はありえない
            //既にひとつ前の評価において、一致しないことが確認済みのため
            
            if (parent.code > target) {
                if (parent.left == null)   //左に進めるか確認
                    return -1;             //該当なし
                else {
                    if (parent.left.code == target) { //左ノードは一致するか
                        //一致したら削除、その親のポインタも修正
                        parent.left = DeleteAndNewSubTree(parent.left);
                        return 0;
                    }
                    else {
                        //一致しなったら、単に左に進む
                        parent = parent.left;
                        continue;
                    }
                }
            }
            else {
                if (parent.right == null)  //右に進めるか確認
                    return -1;             //該当なし
                else {
                    if (parent.right.code == target) { //右ノードは一致するか
                        //一致したら削除、その親のポインタも修正
                        parent.right = DeleteAndNewSubTree(parent.right);
                        return 0;
                    }
                    else {
                        //一致しなったら、単に右に進む
                        parent = parent.right;
                        continue;
                    }
                }
            }
        }
    }

    //特定ノードを削除し、その空きノードを埋めるように、
    //削除ノードの配下にある子ノードのいずれかを昇格させて再構築する
    //再構築によって残ったツリーを返す
    //実際に返るのは、新たに昇格してた子ノードをトップとするサブツリー
    private static Node DeleteAndNewSubTree(Node p) {
        if (p.left == null) {
            if (p.right == null) {
                //両方の子ノードがない場合、返すべきツリーはもうないので null を返す
                return null;
            }
            else {
                //右だけある場合、右を昇格させるため返す(右が新しいサブツリーのトップ)
                return root.right;
            }
        }
        else {
            if (root.right == null) {
                //左だけある場合、左を昇格させるため返す(右が新しいサブツリーのトップ)
                return root.left;
            }
            else {
                //両方の子ノードがある場合、左サブツリーの最大値を昇格させる
                Node right = p.right;              //現在の右を退避
                p = ExtractMaxAndRebuild(p.left);  //左サブツリーの最大値を昇格する
                p.right = right;                   // 退避していた右サブツリーを再設定
                return p;
            }
        }
    }


    //特定ノードを基準とした(トップとした)サブツリーの中で、最大値のノードを取り出し、
    //それがトップになるよう、既存のトップの親になるよう再構築する
    //新しいトップを戻り値として返す
    //なお、最大値であるノードがトップになったので、そのトップの右は空きノードとなる
    private static Node ExtractMaxAndRebuild(Node r) {
        Node p, parent;
        for (p = r, parent = null;
             p.right != null; parent = p, p = p.right) ;  // 右の空きが見つかるまで進める
        // 進む際に、親ノードを退避する
        //この時点で最大値はノード p
        //その親は、parent

        if (parent != null) {      //親があれば親から削除し、自分はトップよりもさらに親になる
                                   //親がなければ、そもそも p がツリーのトップなので何もしない
            parent.right = null;
            p.left = r;
        }

        return p;                 //最大値 p を返す
    }
    //==== ==== ★★ ★★ ==== ==== 
}
