﻿using System;

class HSortApp
{
    static void Main(string[] args)
    {
        int[] dat = { 22, 65, 72, 17, 48, 50, 39, 90, 80 };

        for (int i = 0; i < dat.Length; i++)
            Console.Write("{0,4}", dat[i]);
        Console.WriteLine();

        HeapSort(dat, dat.Length);  //ヒープソート

        for (int i = 0; i < dat.Length; i++)
            Console.Write("{0,4}", dat[i]);
        Console.WriteLine();
    }

    //  以下のコード断片は、C# コンパイラと Java コンパイラのどちらでも有効な実装です。 

    //==== ==== ★★ヒープソート本体★★ ==== ====
    //  [] dat ---- ソート対象の配列本体
    //  n      ---- 要素の数(2以上であることが前提)
    private static void HeapSort(int[] dat, int n) {
        //初期ヒープの作成
        InitHeap(dat, n);
        //トップを交換しながら、ヒープを小さくしつつ、ヒープを修復
        for (int size = n; size >= 2; ) {
            // トップの値を配列の末尾と交換
            int w = dat[size - 1];  //配列の添え字はゼロから始まるので補正
            dat[size - 1] = dat[0];
            dat[0] = w;
            // ヒープを一つ小さくして修復
            size--;
            RebuildHeap(dat, 0, size - 1);
        }
    }

    //初期ヒープの作成
    //  [] dat ---- ソート対象の配列本体
    //  n      ---- ヒープとして使用される要素数
    private static void InitHeap(int[] dat, int n) {
        for (int k = n / 2; k >= 1; k--) {
            // 分かりやすくするため、各計算式には多少の冗長性があります

            // 親と左右の子の位置を求める(添え字は0から始まるので 1減算)
            int parent = k - 1;           // 親
            int left = (k * 2) - 1;       // 左の子
            int right = (k * 2 + 1) - 1;  // 右の子

            // 親と二つの子で比較
            int winner = Battle(dat, parent, left, right, n - 1);
            if (parent == winner) //親が勝者ならば何もせずに、次のノードの評価へ
                continue;
            else {                //子が勝者ならば交換
                // dat[parent] と dat[winner] を交換
                int w = dat[parent];
                dat[parent] = dat[winner];
                dat[winner] = w;
                // 交換され下位に落ちた新たな親の位置(winner)をトップとして、ヒープを修復
                RebuildHeap(dat, winner, n - 1);
            }
        }
    }
    
    //ヒープを指定したトップから修復
    //  [] dat ---- ソート対象の配列本体
    //  top    ---- 修復したいトップの位置(0から始まる配列の添え字)
    //  bottom ---- ヒープとして使用する範囲の終端(0から始まる配列の添え字)
    private static void RebuildHeap(int[] dat, int top, int bottom) {
        int k = top + 1;          //図に合わせるために、1から始まる変数 k を使用
        int size = bottom + 1;    //図に合わせるために、1から始まるヒープ終端を表す変数 size を使用
        int parent, left, right;  //親の位置、左の子の位置、右の子の位置 (どれも添え字が 0 から始まることが前提)

        // 親と左右の子の位置を求める(添え字は0から始まるので 1減算)
        parent = k - 1;
        left = (k * 2) - 1;
        right = (k * 2 + 1) - 1;

        // 親と二つの子で比較
        int winner = Battle(dat, parent, left, right, bottom);
        if (parent == winner) //親が勝者ならば何もせずに終わる
            return;
        else {                //子が勝者ならば交換
            int w = dat[parent];
            dat[parent] = dat[winner];
            dat[winner] = w;
            // 交換され下位に落ちた新たな親の位置をトップとして、ヒープを修復
            RebuildHeap(dat, winner, size - 1);
        }
    }

    //親と二つの子を評価し、勝者の位置を返す(配列の添え字)
    //  [] dat        ---- ソート対象の配列本体
    //  parent        ---- 親の位置
    //  left と right ---- 左右の子の位置
    //  bottom        ---- ヒープとして使用する範囲の終端
    private static int Battle(int[] dat, int parent, int left, int right, int bottom) {
        if (left > bottom)  //左の位置がヒープ外ならば、そもそも両方の子はないので、親の勝ち
            return parent;

        // この時点で少なくとも左の子は存在する

        // 左右の子の対決
        int childWinner;    //左右のうちの勝者の添え字
        if (right > bottom)                //右の位置がヒープ外ならば、左の子の勝ち
            childWinner = left;
        else if (dat[left] >= dat[right])  //両方の子があるなら、左右を比較
            childWinner = left;            //左の勝ち
        else
            childWinner = right;           //右の勝ち

        // 最後に親と対決
        if (dat[parent] >= dat[childWinner])
            return parent;       // 親の勝ち
        else
            return childWinner;  // 勝者の子の勝ち
    }
    //==== ==== ★★ ★★ ==== ==== 
}
